package net.sf.cpsolver.coursett.criteria;

import net.sf.cpsolver.coursett.Constants;
import net.sf.cpsolver.coursett.constraint.RoomConstraint;
import net.sf.cpsolver.coursett.model.TimeLocation;
import net.sf.cpsolver.ifs.util.DataProperties;

/**
 * Useless half-hours. This criterion counts cases when there is an empty half hour in a room.
 * Such half-hours should be generally avoided as usually any class takes more than half an hour.  
 * <br>
 * 
 * @version CourseTT 1.2 (University Course Timetabling)<br>
 *          Copyright (C) 2006 - 2011 Tomas Muller<br>
 *          <a href="mailto:muller@unitime.org">muller@unitime.org</a><br>
 *          <a href="http://muller.unitime.org">http://muller.unitime.org</a><br>
 * <br>
 *          This library is free software; you can redistribute it and/or modify
 *          it under the terms of the GNU Lesser General Public License as
 *          published by the Free Software Foundation; either version 3 of the
 *          License, or (at your option) any later version. <br>
 * <br>
 *          This library is distributed in the hope that it will be useful, but
 *          WITHOUT ANY WARRANTY; without even the implied warranty of
 *          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *          Lesser General Public License for more details. <br>
 * <br>
 *          You should have received a copy of the GNU Lesser General Public
 *          License along with this library; if not see
 *          <a href='http://www.gnu.org/licenses/'>http://www.gnu.org/licenses/</a>.
 */
public class UselessHalfHours extends BrokenTimePatterns {

    @Override
    public double getWeightDefault(DataProperties config) {
        return Constants.sPreferenceLevelStronglyDiscouraged * config.getPropertyDouble("Comparator.UselessSlotWeight", 0.1);
    }
    
    @Override
    public String getPlacementSelectionWeightName() {
        return "Placement.UselessSlotsWeight";
    }
    
    @Override
    protected int penalty(RoomConstraint rc) {
        return countUselessSlotsHalfHours(rc);
    }

    @Override
   protected int penalty(RoomConstraint rc, TimeLocation value) {
        return countUselessSlotsHalfHours(rc, value);
    }
    
    private static boolean isUselessBefore(RoomConstraint rc, int slot) {
        int s = slot % Constants.SLOTS_PER_DAY;
        if (s - 1 < 0 || s + 6 >= Constants.SLOTS_PER_DAY)
            return false;
        return (!rc.getResource(slot - 1).isEmpty() &&
                rc.getResource(slot + 0).isEmpty() &&
                rc.getResource(slot + 1).isEmpty() &&
                rc.getResource(slot + 2).isEmpty() &&
                rc.getResource(slot + 3).isEmpty() &&
                rc.getResource(slot + 4).isEmpty() &&
                rc.getResource(slot + 5).isEmpty());
    }

    private static boolean isUselessAfter(RoomConstraint rc, int slot) {
        int s = slot % Constants.SLOTS_PER_DAY;
        if (s - 1 < 0 || s + 6 >= Constants.SLOTS_PER_DAY)
            return false;
        return (rc.getResource(slot + 0).isEmpty() &&
                rc.getResource(slot + 1).isEmpty() &&
                rc.getResource(slot + 2).isEmpty() &&
                rc.getResource(slot + 3).isEmpty() &&
                rc.getResource(slot + 4).isEmpty() &&
                rc.getResource(slot + 5).isEmpty() &&
                !rc.getResource(slot + 6).isEmpty());
    }

    /** Number of useless half hours for this room */
    protected static int countUselessSlotsHalfHours(RoomConstraint rc, TimeLocation time) {
        int ret = 0;
        int slot = time.getStartSlot() % Constants.SLOTS_PER_DAY;
        int days = time.getDayCode();
        for (int d = 0; d < Constants.NR_DAYS; d++) {
            if ((Constants.DAY_CODES[d] & days) == 0)
                continue;
            if (isUselessBefore(rc, d * Constants.SLOTS_PER_DAY + slot - 6))
                ret ++;
            if (isUselessAfter(rc, d * Constants.SLOTS_PER_DAY + slot + time.getNrSlotsPerMeeting()))
                ret ++;
        }
        return ret;
    }
    
    private static boolean isUseless(RoomConstraint rc, int slot) {
        int s = slot % Constants.SLOTS_PER_DAY;
        if (s - 1 < 0 || s + 6 >= Constants.SLOTS_PER_DAY)
            return false;
        return (!rc.getResource(slot - 1).isEmpty() &&
                rc.getResource(slot + 0).isEmpty() &&
                rc.getResource(slot + 1).isEmpty() &&
                rc.getResource(slot + 2).isEmpty() &&
                rc.getResource(slot + 3).isEmpty() &&
                rc.getResource(slot + 4).isEmpty() &&
                rc.getResource(slot + 5).isEmpty() &&
                !rc.getResource(slot + 6).isEmpty());
    }

    /** Number of useless slots for this room */
    public static int countUselessSlotsHalfHours(RoomConstraint rc) {
        int ret = 0;
        for (int d = 0; d < Constants.NR_DAYS; d++) {
            for (int s = 0; s < Constants.SLOTS_PER_DAY; s++) {
                int slot = d * Constants.SLOTS_PER_DAY + s;
                if (isUseless(rc, slot))
                    ret++;
            }
        }
        return ret;
    }
}
